#  Copyright 2021-2025 ZBW – Leibniz Information Centre for Economics
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Parse command line args and call respective interface."""

import argparse
import json
import logging
import logging.config
from pathlib import Path

from qualle.interface.config import (
    TrainSettings,
    SubthesauriLabelCalibrationSettings,
    RegressorSettings,
    EvalSettings,
    FeaturesEnum,
    RESTSettings,
    PredictSettings,
)
from qualle.interface.internal import train, evaluate, predict
from qualle.interface.rest import run
from qualle.utils import get_logger

PATH_TO_MODEL_FILE_STR = "Path to model file"


def config_logging(config_file=None, debug=False):
    logger = get_logger()
    if config_file:
        logging.config.fileConfig(config_file)
    else:
        ch = logging.StreamHandler()
        ch.setLevel(logging.DEBUG if debug else logging.INFO)
        logger.addHandler(ch)
        logger.setLevel(logging.DEBUG if debug else logging.INFO)


class CliValidationError(Exception):
    pass


def handle_train(args: argparse.Namespace):
    logger = get_logger()

    if args.slc:
        if not all((args.thsys, args.s_type, args.c_type, args.c_uri_prefix)):
            raise CliValidationError(
                "Not all required arguments for Subthesauri Label Calibration "
                "have been passed. Please see usage."
            )
        if args.subthesauri:
            subthesauri = args.subthesauri[0].split(",")
        else:
            subthesauri = []

        slc = SubthesauriLabelCalibrationSettings(
            thesaurus_file=args.thsys[0],
            subthesaurus_type=args.s_type[0],
            concept_type=args.c_type[0],
            concept_type_prefix=args.c_uri_prefix[0],
            subthesauri=subthesauri,
            use_sparse_count_matrix=args.use_sparse_count_matrix,
        )
    else:
        slc = None

    features = []
    if args.features:
        cli_features = set(args.features)
        if {"all", "confidence"} & cli_features:
            features.append(FeaturesEnum.CONFIDENCE)
        if {"all", "text"} & cli_features:
            features.append(FeaturesEnum.TEXT)
        logger.debug(
            "Use features in addition to LabelCalibration: %s",
            list(map(lambda f: f.value, features)),
        )

    lc_regressor_config = json.loads(args.label_calibrator_regressor[0])
    lc_regressor = RegressorSettings(
        regressor_class=lc_regressor_config["class"],
        params={k: v for k, v in lc_regressor_config.items() if k != "class"},
    )
    qe_regressor_config = json.loads(args.quality_estimator_regressor[0])
    qe_regressor = RegressorSettings(
        regressor_class=qe_regressor_config["class"],
        params={k: v for k, v in qe_regressor_config.items() if k != "class"},
    )

    settings = TrainSettings(
        label_calibrator_regressor=lc_regressor,
        quality_estimator_regressor=qe_regressor,
        train_data_path=args.train_data_path,
        output_path=args.output,
        features=features,
        subthesauri_label_calibration=slc,
        should_debug=args.should_debug,
    )

    train(settings)


def handle_eval(args: argparse.Namespace):
    settings = EvalSettings(test_data_path=args.test_data_path, mdl_file=args.model)
    evaluate(settings)


def handle_rest(args: argparse.Namespace):
    settings = RESTSettings(mdl_file=args.model, port=args.port[0], host=args.host[0])
    run(settings)


def handle_predict(args: argparse.Namespace):
    predict_data_path = args.predict_data_path
    output_path = args.output[0] if args.output else None

    if predict_data_path.is_file() and not output_path:
        raise CliValidationError(
            "output file has to be specified if tsv file has been specified"
        )
    settings = PredictSettings(
        predict_data_path=predict_data_path,
        mdl_file=args.model,
        output_path=output_path,
    )
    predict(settings)


def add_eval_parser(subparsers: argparse._SubParsersAction):
    eval_parser = subparsers.add_parser(
        "eval", description="Run evaluation on training data."
    )
    eval_parser.set_defaults(func=handle_eval)

    eval_parser.add_argument(
        "test_data_path",
        type=str,
        help="Path to test data. "
        "Accepted inputs are either a tsv file or "
        "a folder in Annif format.",
    )
    eval_parser.add_argument("model", type=str, help=PATH_TO_MODEL_FILE_STR)


def add_slc_group(parsers: argparse.ArgumentParser):
    slc_group = parsers.add_argument_group(
        "subthesauri-label-calibration",
        description="Run Label Calibration by distinguishing between label "
        "count for different subthesauri. "
        "All Arguments in this group ,"
        " except --subthesauri , are required. "
        "If --subthesauri is not passed, subthesauri used "
        "are automatically detected by type.",
    )
    slc_group.add_argument(
        "--slc", action="store_true", help="Activate Subthesauri Label Calibration."
    )
    slc_group.add_argument(
        "--thsys",
        type=str,
        nargs=1,
        help="path to thesaurus file (must be in RDF format)",
    )
    slc_group.add_argument(
        "--s-type",
        type=str,
        nargs=1,
        help="subthesaurus type uri, e.g.:  "
        "http://zbw.eu/namespaces/zbw-extensions/Thsys",
    )
    slc_group.add_argument(
        "--c-type",
        type=str,
        nargs=1,
        help="concept type uri, e.g.: "
        "http://zbw.eu/namespaces/zbw-extensions/Descriptor",
    )
    slc_group.add_argument(
        "--c-uri-prefix",
        type=str,
        nargs=1,
        help="concept uri prefix, e.g.: http://zbw.eu/stw/descriptor)",
    )
    slc_group.add_argument(
        "--subthesauri",
        type=str,
        nargs=1,
        help="subthesauri URIs as comma-separated list, "
        "e.g.: http://zbw.eu/stw/thsys/v,"
        "http://zbw.eu/stw/thsys/b,"
        "http://zbw.eu/stw/thsys/n",
    )
    slc_group.add_argument(
        "--use-sparse-count-matrix",
        action="store_true",
        help="Use matrix in SciPy Sparse format to store count of labels per"
        " Subthesauri in memory. Useful to reduce memory usage when"
        " mapping labels to subthesauri "
        " (used for label calibration training and"
        "  for calculating the input of the recall predictor)"
        " if the distribution of labels over Subthesauri is sparse.",
    )


def add_train_parser(subparsers: argparse._SubParsersAction):
    train_parser = subparsers.add_parser(
        "train", description="Run training using default estimator."
    )
    train_parser.set_defaults(func=handle_train)
    train_parser.add_argument(
        "train_data_path",
        type=str,
        help="Path to train data. "
        "Accepted inputs are either a tsv file or "
        "a folder in Annif format.",
    )
    train_parser.add_argument("output", type=str, help="Path to output model file")
    train_parser.add_argument(
        "--features",
        "-f",
        choices=["all", "confidence", "text"],
        action="append",
        type=str,
        help="Use features in addition to Label Calibration. "
        "Can be passed multiple times. "
        'If "all" is passed, all features will be used.',
    )
    train_parser.add_argument(
        "--label-calibrator-regressor",
        type=str,
        nargs=1,
        help="Specifiy regressor to use for Label Calibration in JSON format."
        'Requires property "class" with fully qualified name of the '
        "scikit-learn regressor class to use  for the Label Calibrator. "
        'E.g.: {"class": "sklearn.ensemble.GradientBoostingRegressor",'
        '"min_samples_leaf": 30, "max_depth": 5, "n_estimators": 10} ',
        default=[
            '{"class": "sklearn.ensemble.GradientBoostingRegressor",'
            '"min_samples_leaf": 30, "max_depth": 5, "n_estimators": 10}'
        ],
    )
    train_parser.add_argument(
        "--quality-estimator-regressor",
        type=str,
        nargs=1,
        help="Specifiy regressor to use for Label Calibration in JSON format."
        'Requires property "class" with fully qualified name of the '
        "scikit-learn regressor class to use  for the Quality Estimator. "
        'E.g.: {"class": "sklearn.ensemble.GradientBoostingRegressor",'
        '"min_samples_leaf": 30, "max_depth": 5, "n_estimators": 10} ',
        default=[
            '{"class": "sklearn.ensemble.GradientBoostingRegressor",'
            '"n_estimators": 10, "max_depth": 8}'
        ],
    )
    add_slc_group(train_parser)


def add_rest_parser(subparsers: argparse._SubParsersAction):
    rest_parser = subparsers.add_parser(
        "rest", description="Run Qualle as REST Webservice with predict endpoint"
    )
    rest_parser.add_argument("model", type=str, help=PATH_TO_MODEL_FILE_STR)
    rest_parser.add_argument(
        "--port", "-p", type=int, nargs=1, help="Port to listen on", default=[8000]
    )
    rest_parser.add_argument(
        "--host", type=str, nargs=1, help="Host to listen on", default=["127.0.0.1"]
    )
    rest_parser.set_defaults(func=handle_rest)


def add_predict_parser(subparsers: argparse._SubParsersAction):
    predict_parser = subparsers.add_parser(
        "predict",
        description="Predict the quality for a collection of automated "
        "subject indexing (MLC) results.",
    )
    predict_parser.set_defaults(func=handle_predict)

    predict_parser.add_argument(
        "predict_data_path",
        type=Path,
        help="Path to data. "
        "Accepted inputs are either a tsv file or a folder in annif "
        "format. If a tsv file is provided, the flag --output must be "
        "specified. If a folder is provided, the predicted quality is "
        'written into files using the ".qualle" prefix inside the folder.',
    )
    predict_parser.add_argument("model", type=str, help=PATH_TO_MODEL_FILE_STR)
    predict_parser.add_argument(
        "--output",
        type=Path,
        nargs=1,
        help="Output file to write the quality estimation. Only used if a tsv "
        "file is specified as input. The quality estimation for each "
        "line of the tsv file is written to the output on a separate "
        "line, preserving the order in the tsv file.",
    )


def cli_entrypoint():
    parser = argparse.ArgumentParser(
        description="Quality Estimation for Automatic Subject Indexing"
    )

    parser.add_argument(
        "--debug",
        action="store_true",
        help="Set Log Level to Debug",
        dest="should_debug",
    )
    parser.add_argument(
        "--logging-conf",
        nargs=1,
        type=str,
        help="Path to logging config file in configparser format. "
        'The name of the logger which has to be configured is "qualle"',
    )
    subparsers = parser.add_subparsers(
        title="Subcommands", required=True, dest="command"
    )

    add_eval_parser(subparsers)
    add_train_parser(subparsers)
    add_rest_parser(subparsers)
    add_predict_parser(subparsers)

    args = parser.parse_args()

    config_logging(config_file=args.logging_conf, debug=args.should_debug)
    args.func(args)



#  Copyright 2021-2025 ZBW – Leibniz Information Centre for Economics
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from sklearn.feature_extraction.text import CountVectorizer
from sklearn.pipeline import FeatureUnion

from qualle.features.text.count import (
    CountCharsFeature,
    CountWordsFeature,
    CountSpecialCharsFeature,
    CountUpperCharsFeature,
    CountDigitsFeature,
)

_NAME_VECTOR_FEATURE = "vectorizer"
_NAME_CHAR_FEATURE = "n_chars"
_NAME_WORD_FEATURE = "n_words"
_NAME_SPECIAL_CHARS_FEATURE = "n_special"
_NAME_UPPER_FEATURE = "n_upper"
_NAME_DIGIT_FEATURE = "n_digits"


class TextFeatures(FeatureUnion):
    """Features based on the text of a document"""

    def __init__(self):
        super().__init__(
            [
                (_NAME_VECTOR_FEATURE, CountVectorizer(lowercase=False)),
                (_NAME_CHAR_FEATURE, CountCharsFeature()),
                (_NAME_WORD_FEATURE, CountWordsFeature()),
                (_NAME_SPECIAL_CHARS_FEATURE, CountSpecialCharsFeature()),
                (_NAME_UPPER_FEATURE, CountUpperCharsFeature()),
                (_NAME_DIGIT_FEATURE, CountDigitsFeature()),
            ]
        )



#  Copyright 2021-2025 ZBW – Leibniz Information Centre for Economics
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import logging
from contextlib import contextmanager
from time import perf_counter
from typing import List

from qualle.models import Labels


def recall(true_labels: List[Labels], predicted_labels: List[Labels]) -> List:
    return [
        len(set(tc) & set(pc)) / len_tc if (len_tc := len(tc)) > 0 else 0
        for tc, pc in zip(true_labels, predicted_labels)
    ]


def get_logger():
    return logging.getLogger("qualle")


@contextmanager
def timeit():
    """Context manager to time the code block wrapped by the manager.

    Yielded object is a lambda function which can be called to compute the
    duration (in seconds) of the executed code block.
    """
    start = perf_counter()
    yield lambda: perf_counter() - start



#  Copyright 2021-2025 ZBW  – Leibniz Information Centre for Economics
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from collections import namedtuple
from dataclasses import dataclass
from pathlib import Path
from typing import List, Tuple, Iterable, Any

from qualle.models import TrainData, PredictData

DocumentId = str
Score = float
DocumentQualityEstimation = Tuple[Any, DocumentId]


@dataclass
class AnnifTrainData:
    document_ids: List[DocumentId]
    train_data: TrainData


@dataclass
class AnnifPredictData:
    document_ids: List[DocumentId]
    predict_data: PredictData


_Data = namedtuple(
    "data", ["docs", "predicted_labels", "scores", "true_labels", "doc_ids"]
)


class AnnifHandler:
    def __init__(self, dir: Path):
        self._dir = dir

    def load_train_input(self) -> AnnifTrainData:
        return self._map_to_annif_train_data(self._input_from_annif())

    def load_predict_input(self) -> AnnifPredictData:
        return self._map_to_annif_predict_data(
            self._input_from_annif(include_true_labels=False)
        )

    def store_quality_estimations(
        self, quality_ests: Iterable[DocumentQualityEstimation]
    ):
        for quality_est, doc_id in quality_ests:
            score_fp = self._dir / (doc_id + ".qualle")
            score_fp.write_text(str(quality_est))

    @staticmethod
    def _map_to_annif_train_data(data: _Data) -> AnnifTrainData:
        return AnnifTrainData(
            train_data=TrainData(
                predict_data=AnnifHandler._map_to_predict_data(data),
                true_labels=data.true_labels,
            ),
            document_ids=data.doc_ids,
        )

    @staticmethod
    def _map_to_annif_predict_data(data: _Data) -> AnnifPredictData:
        return AnnifPredictData(
            predict_data=AnnifHandler._map_to_predict_data(data),
            document_ids=data.doc_ids,
        )

    @staticmethod
    def _map_to_predict_data(data: _Data) -> PredictData:
        return PredictData(
            docs=data.docs, predicted_labels=data.predicted_labels, scores=data.scores
        )

    def _input_from_annif(self, include_true_labels=True) -> _Data:
        docs = []
        pred_labels = []
        true_labels = []
        scores = []
        doc_ids = []
        for pred_pth in self._dir.glob("*.annif"):
            doc_id = pred_pth.with_suffix("").name
            doc_ids.append(doc_id)
            with pred_pth.open() as fp:
                scores_for_doc = []
                pred_labels_for_doc = []
                for line in fp.readlines():
                    split = line.rstrip("\n").split("\t")
                    concept_id = self._extract_concept_id_from_annif_label(split[0])
                    pred_labels_for_doc.append(concept_id)
                    score = float(split[2])
                    scores_for_doc.append(score)
                pred_labels.append(pred_labels_for_doc)
                scores.append(scores_for_doc)
            doc_pth = pred_pth.with_suffix(".txt")
            with doc_pth.open() as fp:
                docs.append("".join(list(fp.readlines())))
            label_pth = pred_pth.with_suffix(".tsv")
            if include_true_labels and label_pth.exists():
                with label_pth.open() as fp:
                    true_labels_for_doc = []
                    for line in fp.readlines():
                        split = line.rstrip("\n").split("\t")
                        concept_id = self._extract_concept_id_from_annif_label(split[0])
                        true_labels_for_doc.append(concept_id)
                    true_labels.append(true_labels_for_doc)
        return _Data(
            docs=docs,
            predicted_labels=pred_labels,
            scores=scores,
            true_labels=true_labels,
            doc_ids=doc_ids,
        )

    @staticmethod
    def _extract_concept_id_from_annif_label(uri: str) -> str:
        split = uri.split("/")
        # remove '>' at end of URI
        return split[-1][:-1]



#  Copyright 2021-2025 ZBW – Leibniz Information Centre for Economics
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from typing import Set, Type, List

from sklearn.base import RegressorMixin

from qualle.features.base import Features
from qualle.features.combined import CombinedFeaturesData, CombinedFeatures
from qualle.features.confidence import ConfidenceFeatures
from qualle.features.label_calibration.base import AbstractLabelCalibrator
from qualle.features.label_calibration.simple_label_calibration import (
    SimpleLabelCalibrationFeatures,
)
from qualle.features.label_calibration.thesauri_label_calibration import (
    ThesauriLabelCalibrationFeatures,
)
from qualle.features.text import TextFeatures
from qualle.models import PredictData, LabelCalibrationData
from qualle.pipeline import QualityEstimationPipeline
from qualle.quality_estimation import QualityEstimator

FeaturesTypes = Set[Type[Features]]


class Trainer:
    def __init__(
        self,
        train_data,
        label_calibrator: AbstractLabelCalibrator,
        quality_regressor: RegressorMixin,
        features: List[Features],
        should_debug=False,
    ):
        combined_features = CombinedFeatures(features)
        rp = QualityEstimator(regressor=quality_regressor, features=combined_features)
        self._qe_p = QualityEstimationPipeline(
            recall_predictor=rp,
            label_calibrator=label_calibrator,
            should_debug=should_debug,
            features_data_mapper=FeaturesDataMapper(set(f.__class__ for f in features)),
        )
        self._train_data = train_data

    def train(self) -> QualityEstimationPipeline:
        self._qe_p.train(self._train_data)

        return self._qe_p


class FeaturesDataMapper:
    def __init__(self, features_types: FeaturesTypes):
        self._features_types = features_types

    def __call__(
        self, p_data: PredictData, l_data: LabelCalibrationData
    ) -> CombinedFeaturesData:
        features_data = dict()
        for ftype in self._features_types:
            if ftype == ConfidenceFeatures:
                features_data[ConfidenceFeatures] = p_data.scores
            if ftype == TextFeatures:
                features_data[TextFeatures] = p_data.docs
            if ftype in (
                SimpleLabelCalibrationFeatures,
                ThesauriLabelCalibrationFeatures,
            ):
                features_data[ftype] = l_data
        return features_data



#  Copyright 2021-2025 ZBW – Leibniz Information Centre for Economics
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from contextlib import contextmanager
from typing import List, Callable, Any, Collection

from sklearn.model_selection import cross_val_predict

from qualle.features.label_calibration.base import AbstractLabelCalibrator
from qualle.models import TrainData, PredictData, LabelCalibrationData
from qualle.quality_estimation import QualityEstimator
from qualle.utils import recall, get_logger, timeit


class QualityEstimationPipeline:
    def __init__(
        self,
        label_calibrator: AbstractLabelCalibrator,
        recall_predictor: QualityEstimator,
        features_data_mapper: Callable[[PredictData, LabelCalibrationData], Any],
        should_debug=False,
    ):
        self._label_calibrator = label_calibrator
        self._recall_predictor = recall_predictor
        self._features_data_mapper = features_data_mapper
        self._should_debug = should_debug
        self._logger = get_logger()

    def train(self, data: TrainData):
        predict_data = data.predict_data
        with self._debug("cross_val_predict with label calibrator"):
            predicted_no_of_labels = cross_val_predict(
                self._label_calibrator, predict_data.docs, data.true_labels
            )

        with self._debug("label calibrator fit"):
            self._label_calibrator.fit(predict_data.docs, data.true_labels)

        label_calibration_data = LabelCalibrationData(
            predicted_labels=predict_data.predicted_labels,
            predicted_no_of_labels=predicted_no_of_labels,
        )
        features_data = self._features_data_mapper(predict_data, label_calibration_data)
        with self._debug("recall computation"):
            true_recall = recall(data.true_labels, predict_data.predicted_labels)

        with self._debug("RecallPredictor fit"):
            self._recall_predictor.fit(features_data, true_recall)

    def predict(self, data: PredictData) -> List[float]:
        zero_idxs = self._get_pdata_idxs_with_zero_labels(data)
        data_with_labels = self._get_pdata_with_labels(data, zero_idxs)
        if data_with_labels.docs:
            predicted_no_of_labels = self._label_calibrator.predict(
                data_with_labels.docs
            )
            label_calibration_data = LabelCalibrationData(
                predicted_labels=data_with_labels.predicted_labels,
                predicted_no_of_labels=predicted_no_of_labels,
            )
            features_data = self._features_data_mapper(
                data_with_labels, label_calibration_data
            )
            predicted_recall = self._recall_predictor.predict(features_data)
            recall_scores = self._merge_zero_recall_with_predicted_recall(
                predicted_recall=predicted_recall,
                zero_labels_idx=zero_idxs,
            )
        else:
            recall_scores = [0] * len(data.predicted_labels)
        return recall_scores

    @staticmethod
    def _get_pdata_idxs_with_zero_labels(data: PredictData) -> Collection[int]:
        return [
            i for i in range(len(data.predicted_labels)) if not data.predicted_labels[i]
        ]

    @staticmethod
    def _get_pdata_with_labels(
        data: PredictData, zero_labels_idxs: Collection[int]
    ) -> PredictData:
        non_zero_idxs = [
            i for i in range(len(data.predicted_labels)) if i not in zero_labels_idxs
        ]
        return PredictData(
            docs=[data.docs[i] for i in non_zero_idxs],
            predicted_labels=[data.predicted_labels[i] for i in non_zero_idxs],
            scores=[data.scores[i] for i in non_zero_idxs],
        )

    @staticmethod
    def _merge_zero_recall_with_predicted_recall(
        predicted_recall: List[float],
        zero_labels_idx: Collection[int],
    ):
        recall_scores = []
        j = 0
        for i in range(len(zero_labels_idx) + len(predicted_recall)):
            if i in zero_labels_idx:
                recall_scores.append(0)
            else:
                recall_scores.append(predicted_recall[j])
                j += 1
        return recall_scores

    @contextmanager
    def _debug(self, method_name):
        if self._should_debug:
            with timeit() as t:
                yield
            self._logger.debug("Ran %s in %.4f seconds", method_name, t())
        else:
            yield

    def __str__(self):
        return f"{self._label_calibrator}\n{self._recall_predictor}"



#  Copyright 2021-2025 ZBW – Leibniz Information Centre for Economics
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from typing import List, Dict

import numpy as np
from sklearn.metrics import explained_variance_score, mean_squared_error

from qualle.models import TrainData
from qualle.pipeline import QualityEstimationPipeline
from qualle.utils import recall


class Evaluator:
    def __init__(self, test_data: TrainData, qe_p: QualityEstimationPipeline):
        self._test_data = test_data
        self._qe_p = qe_p

    def evaluate(self) -> Dict:
        predict_data = self._test_data.predict_data
        pred_recall = self._qe_p.predict(predict_data)
        true_recall = recall(self._test_data.true_labels, predict_data.predicted_labels)

        return scores(true_recall, pred_recall)


def scores(true_recall: List[float], pred_recall: List[float]) -> Dict:
    scores = dict()
    scores["explained_variance_score"] = explained_variance_score(
        true_recall, pred_recall
    )
    scores["mean_squared_error"] = mean_squared_error(true_recall, pred_recall)

    scores["correlation_coefficient"] = np.corrcoef(true_recall, pred_recall)[0, 1]
    return scores



#  Copyright 2021-2025 ZBW – Leibniz Information Centre for Economics
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
from functools import lru_cache
from typing import List, Set, Optional

import numpy as np
from rdflib import URIRef, Graph
from rdflib.namespace import SKOS, RDF
from scipy.sparse import coo_matrix
from sklearn.ensemble import ExtraTreesRegressor
from sklearn.utils.validation import check_is_fitted

from qualle.features.base import Features
from qualle.features.label_calibration.base import AbstractLabelCalibrator
from qualle.label_calibration.category import MultiCategoryLabelCalibrator
from qualle.models import Labels, Documents, LabelCalibrationData, Matrix
from qualle.utils import get_logger


class NotInitializedException(Exception):
    pass


class Thesaurus:
    """Handle access to RDF Graph."""

    def __init__(
        self,
        graph: Graph,
        subthesaurus_type_uri: URIRef,
        concept_type_uri: URIRef,
        concept_uri_prefix: str,
    ):
        self.graph = graph
        self.subthesaurus_type_uri = subthesaurus_type_uri
        self.concept_type_uri = concept_type_uri
        self.concept_uri_prefix_len_ = 1 + len(concept_uri_prefix.rstrip("/"))

    @lru_cache(maxsize=1000)
    def get_concepts_for_subthesaurus(self, subthesaurus: URIRef) -> Set:
        concepts = set()
        for x in self.graph[subthesaurus : SKOS.narrower]:
            if (x, RDF.type, self.subthesaurus_type_uri) in self.graph:
                concepts_from_subthesaurus = self.get_concepts_for_subthesaurus(x)
                concepts = concepts.union(concepts_from_subthesaurus)
            elif (x, RDF.type, self.concept_type_uri) in self.graph:
                concepts.add(self.extract_concept_id_from_uri_ref(x))
            else:
                logging.warning("unknown narrower type %s", str(x))
        return concepts

    def get_all_subthesauri(self) -> List[URIRef]:
        return list(self.graph[: RDF.type : self.subthesaurus_type_uri])

    def extract_concept_id_from_uri_ref(self, concept_uri: URIRef):
        return concept_uri.toPython()[self.concept_uri_prefix_len_ :]


class LabelCountForSubthesauriTransformer:
    """Compute count of labels per Subthesauri for a given RDF Graph."""

    def __init__(
        self,
        use_sparse_count_matrix: bool,
    ):
        self._use_sparse_count_matrix = use_sparse_count_matrix
        self._mapping = None
        self._logger = get_logger()
        self._subthesauri_count = 0

    def init(self, thesaurus: Thesaurus, subthesauri: Optional[List[URIRef]]):
        self._mapping = dict()
        self._subthesauri_count = len(subthesauri)

        for idx, s in enumerate(subthesauri):
            concepts = thesaurus.get_concepts_for_subthesaurus(s)
            for c in concepts:
                if c not in self._mapping:
                    self._mapping[c] = [False] * self._subthesauri_count
                self._mapping[c][idx] = True
        return self

    def transform(self, X: List[Labels]) -> Matrix:
        """Transform rows of concepts to 2-dimensional count matrix

        Each row in the result matrix contains in each column the total
        amount of labels per subthesaurus.
        """
        if not self._mapping:
            raise NotInitializedException()

        if self._use_sparse_count_matrix:
            return self._transform_with_sparse_matrix(X)
        else:
            return self._transform_without_sparse_matrix(X)

    def _transform_with_sparse_matrix(self, X: List[Labels]) -> Matrix:
        values = []
        row_inds = []
        col_inds = []
        for row_idx, row in enumerate(X):
            subthesauri_counts = self._extract_subthesauri_counts(row)
            for col_idx, count in enumerate(subthesauri_counts):
                if count:
                    values.append(count)
                    row_inds.append(row_idx)
                    col_inds.append(col_idx)

        return coo_matrix(
            (values, (row_inds, col_inds)), shape=(len(X), self._subthesauri_count)
        )

    def _transform_without_sparse_matrix(self, X: List[Labels]) -> Matrix:
        count_matrix = np.zeros((len(X), self._subthesauri_count))
        for row_idx, row in enumerate(X):
            subthesauri_counts = self._extract_subthesauri_counts(row)
            count_matrix[row_idx] = subthesauri_counts

        return count_matrix

    def _extract_subthesauri_counts(self, labels: Labels):
        subthesauri_counts = [0] * self._subthesauri_count
        for concept in labels:
            subthesauri_membership = self._mapping.get(concept)
            if not subthesauri_membership:
                self._logger.warning(
                    'Concept "%s" does not appear in any subthesaurus specified'
                    " for label calibration. "
                    "Will ignore this concept for label calibration. ",
                    concept,
                )
            else:
                for j, is_in_subthesauri in enumerate(subthesauri_membership):
                    if is_in_subthesauri:
                        subthesauri_counts[j] = subthesauri_counts[j] + 1
        return subthesauri_counts


class ThesauriLabelCalibrator(AbstractLabelCalibrator):
    def __init__(
        self,
        transformer: LabelCountForSubthesauriTransformer,
        regressor_class=ExtraTreesRegressor,
        regressor_params=None,
    ):
        self.transformer = transformer
        self.regressor_class = regressor_class
        self.regressor_params = regressor_params

    def fit(self, X: Documents, y: List[Labels]):
        self.calibrator_ = MultiCategoryLabelCalibrator(
            regressor_class=self.regressor_class, regressor_params=self.regressor_params
        )
        y_transformed = self.transformer.transform(y)
        self.calibrator_.fit(X, y_transformed)
        return self

    def predict(self, X: Documents):
        check_is_fitted(self)
        return self.calibrator_.predict(X)


class ThesauriLabelCalibrationFeatures(Features):
    def __init__(self, transformer: LabelCountForSubthesauriTransformer):
        self.transformer = transformer

    def transform(self, X: LabelCalibrationData):
        rows = len(X.predicted_no_of_labels)
        no_of_predicted_labels = self.transformer.transform(X.predicted_labels)
        transformed_cols_len = no_of_predicted_labels.shape[1]
        data = np.zeros((rows, 2 * transformed_cols_len))
        data[:, :transformed_cols_len] = X.predicted_no_of_labels
        data[:, transformed_cols_len:] = (
            X.predicted_no_of_labels - no_of_predicted_labels
        )
        return data



#  Copyright 2021-2025 ZBW – Leibniz Information Centre for Economics
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import re
from typing import Callable

import numpy as np

from qualle.features.base import Features


class TextCountFeature(Features):
    """Feature based on the occurence of something in a text."""

    def __init__(self, count_fun: Callable[[str], int]):
        self.count_fun = count_fun

    def fit(self, X=None, y=None):
        return self

    def transform(self, X):
        ret = np.empty((len(X), 1))
        for i, x in enumerate(X):
            ret[i, 0] = self.count_fun(x)
        return ret


class CountCharsFeature(TextCountFeature):
    def __init__(self):
        super().__init__(self._count_char)

    @staticmethod
    def _count_char(txt: str):
        return len(txt)


class CountWordsFeature(TextCountFeature):
    def __init__(self):
        super().__init__(self._count_word)

    @staticmethod
    def _count_word(txt: str):
        return txt.count(" ")


class CountSpecialCharsFeature(TextCountFeature):
    _RE_SPECIAL = re.compile(r"""["'?!()]""")

    def __init__(self):
        super().__init__(self._count_special)

    @classmethod
    def _count_special(cls, txt: str):
        return len(cls._RE_SPECIAL.findall(txt))


class CountUpperCharsFeature(TextCountFeature):
    def __init__(self):
        super().__init__(self._count_upper)

    @staticmethod
    def _count_upper(txt: str):
        return sum((c.isupper() for c in txt))


class CountDigitsFeature(TextCountFeature):
    _RE_DIGIT = re.compile(r"\d")

    def __init__(self):
        super().__init__(self._count_digit)

    @classmethod
    def _count_digit(cls, txt: str):
        return len(cls._RE_DIGIT.findall(txt))



#  Copyright 2021-2025 ZBW – Leibniz Information Centre for Economics
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from typing import List

import numpy as np

from qualle.features.base import Features
from qualle.models import Scores


class ConfidenceFeatures(Features):
    """Features based on aggregating concept level confidence scores."""

    def transform(self, X: List[Scores]):
        _min = [np.min(row) if row else 0.0 for row in X]
        _mean = [np.mean(row) if row else 0.0 for row in X]
        _median = [np.median(row) if row else 0.0 for row in X]
        _prod = [np.prod(row) if row else 0.0 for row in X]
        return np.column_stack([_min, _mean, _median, _prod])



#  Copyright 2021-2025 ZBW – Leibniz Information Centre for Economics
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import List

import numpy as np
from sklearn.base import RegressorMixin
from sklearn.utils.validation import check_is_fitted

from qualle.features.base import Features
from qualle.features.label_calibration.base import AbstractLabelCalibrator
from qualle.label_calibration.simple import LabelCalibrator
from qualle.models import LabelCalibrationData, Documents, Labels


def transform_to_label_count(X: List[Labels]) -> np.array:
    return np.array(list(map(len, X)))


class SimpleLabelCalibrator(AbstractLabelCalibrator):
    def __init__(self, regressor: RegressorMixin):
        self.regressor = regressor

    def fit(self, X: Documents, y: List[Labels]):
        self.calibrator_ = LabelCalibrator(self.regressor)
        y_transformed = transform_to_label_count(y)
        self.calibrator_.fit(X, y_transformed)
        return self

    def predict(self, X: Documents):
        check_is_fitted(self)
        return self.calibrator_.predict(X)


class SimpleLabelCalibrationFeatures(Features):
    def transform(self, X: LabelCalibrationData):
        no_of_pred_labels = transform_to_label_count(X.predicted_labels)
        rows = len(X.predicted_no_of_labels)
        data = np.zeros((rows, 2))
        data[:, 0] = X.predicted_no_of_labels
        data[:, 1] = X.predicted_no_of_labels - no_of_pred_labels
        return data



#  Copyright 2021-2025 ZBW – Leibniz Information Centre for Economics
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.pipeline import Pipeline
from sklearn.utils.validation import check_is_fitted

from qualle.features.base import Features


class QualityEstimator(BaseEstimator, RegressorMixin):
    """Regressor which uses a pipeline with given features and regressor."""

    def __init__(self, regressor: RegressorMixin, features: Features):
        self.regressor = regressor
        self.features = features

    def fit(self, X, y):
        self.pipeline_ = Pipeline(
            [("features", self.features), ("regressor", self.regressor)]
        )
        self.pipeline_.fit(X, y)
        return self

    def predict(self, X):
        check_is_fitted(self)
        return self.pipeline_.predict(X)



#  Copyright 2021-2025 ZBW – Leibniz Information Centre for Economics
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from typing import List, Dict, Any, Type

import numpy as np
import scipy.sparse as sp
from sklearn.utils.validation import check_is_fitted

from qualle.features.base import Features


CombinedFeaturesData = Dict[Type[Features], Any]


class CombinedFeatures(Features):
    """Combine n features by horizontal stacking"""

    def __init__(self, features: List[Features]):
        self.features = features

    def fit(self, X: CombinedFeaturesData, y=None):
        self.features_ = self.features
        for f in self.features_:
            f.fit(X[f.__class__])
        return self

    def transform(self, X: CombinedFeaturesData):
        check_is_fitted(self)
        combined = [f.transform(X[f.__class__]) for f in self.features_]
        issparse = any(map(sp.issparse, combined))
        if issparse:
            return sp.hstack(combined, format="csr")
        else:
            return np.hstack(combined)



#  Copyright 2021-2025 ZBW – Leibniz Information Centre for Economics
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Internal interface to access Qualle functionality"""

from importlib import import_module
from pathlib import Path
from typing import Type, Optional, List, Any, Union

from joblib import dump, load
from rdflib import Graph, URIRef

from qualle.evaluate import Evaluator
from qualle.features.label_calibration.simple_label_calibration import (
    SimpleLabelCalibrator,
    SimpleLabelCalibrationFeatures,
)
from qualle.features.label_calibration.thesauri_label_calibration import (
    ThesauriLabelCalibrator,
    ThesauriLabelCalibrationFeatures,
    LabelCountForSubthesauriTransformer,
    Thesaurus,
)
from qualle.interface.config import TrainSettings, EvalSettings, PredictSettings
from qualle.models import TrainData, PredictData
from qualle.train import Trainer
from qualle.utils import get_logger, timeit
from qualle.interface.data.annif import AnnifHandler
from qualle.interface.data.tsv import (
    load_train_input as load_tsv_train_input,
    load_predict_input,
)


def train(settings: TrainSettings):
    logger = get_logger()
    path_to_train_data = settings.train_data_path
    path_to_model_output_file = settings.output_path
    slc_settings = settings.subthesauri_label_calibration
    features = list(map(lambda f: f.value(), settings.features))

    lc_regressor_cls = _get_class_from_str(
        settings.label_calibrator_regressor.regressor_class
    )
    lc_regressor_params = settings.label_calibrator_regressor.params
    logger.debug(
        "Use (%s %s) as Regressor for Label Calibration",
        settings.label_calibrator_regressor.regressor_class,
        settings.label_calibrator_regressor.params,
    )

    quality_estimator_cls = _get_class_from_str(
        settings.quality_estimator_regressor.regressor_class
    )
    quality_estimator = quality_estimator_cls(
        **settings.quality_estimator_regressor.params
    )
    logger.debug(
        "Use (%s %s) as Regressor for Quality Estimation",
        settings.quality_estimator_regressor.regressor_class,
        settings.quality_estimator_regressor.params,
    )

    with timeit() as t:
        train_data = _load_train_input(path_to_train_data)

    logger.debug("Loaded train data in %.4f seconds", t())

    if slc_settings:
        logger.info("Run training with Subthesauri Label Calibration")
        path_to_graph = slc_settings.thesaurus_file
        g = Graph()
        with timeit() as t:
            g.parse(path_to_graph)
        logger.debug("Parsed RDF Graph in %.4f seconds", t())

        thesaurus = Thesaurus(
            graph=g,
            subthesaurus_type_uri=URIRef(slc_settings.subthesaurus_type),
            concept_type_uri=URIRef(slc_settings.concept_type),
            concept_uri_prefix=slc_settings.concept_type_prefix,
        )
        subthesauri = list(map(lambda s: URIRef(s), slc_settings.subthesauri))
        if not subthesauri:
            logger.info("No subthesauri passed, will extract subthesauri by type")
            subthesauri = thesaurus.get_all_subthesauri()
            logger.debug("Extracted %d subthesauri", len(subthesauri))
        if slc_settings.use_sparse_count_matrix:
            logger.info(
                "Will use Sparse Count Matrix for " "Subthesauri Label Calibration"
            )
        transformer = LabelCountForSubthesauriTransformer(
            use_sparse_count_matrix=slc_settings.use_sparse_count_matrix
        )
        transformer.init(thesaurus=thesaurus, subthesauri=subthesauri)

        label_calibration_features = ThesauriLabelCalibrationFeatures(
            transformer=transformer
        )
        features.append(label_calibration_features)
        t = Trainer(
            train_data=train_data,
            label_calibrator=ThesauriLabelCalibrator(
                regressor_class=lc_regressor_cls,
                regressor_params=lc_regressor_params,
                transformer=transformer,
            ),
            quality_regressor=quality_estimator,
            features=features,
            should_debug=settings.should_debug,
        )
    else:
        logger.info("Run training with Simple Label Calibration")
        label_calibration_features = SimpleLabelCalibrationFeatures()
        features.append(label_calibration_features)
        t = Trainer(
            train_data=train_data,
            label_calibrator=SimpleLabelCalibrator(
                lc_regressor_cls(**lc_regressor_params)
            ),
            quality_regressor=quality_estimator,
            features=features,
            should_debug=settings.should_debug,
        )

    model = t.train()
    logger.info("Store trained model in %s", path_to_model_output_file)
    dump(model, path_to_model_output_file)


def _get_class_from_str(fully_qualified_path: str) -> Type:
    split = fully_qualified_path.split(".")
    module = ".".join(split[:-1])
    cls_name = split[-1]
    return getattr(import_module(module), cls_name)


def evaluate(settings: EvalSettings):
    logger = get_logger()
    path_to_test_data = settings.test_data_path
    path_to_model_file = settings.mdl_file
    model = load_model(str(path_to_model_file))
    logger.info("Run evaluation with model:\n%s", model)
    test_input = _load_train_input(path_to_test_data)
    ev = Evaluator(test_input, model)
    eval_data = ev.evaluate()
    logger.info("\nScores:")
    for metric, score in eval_data.items():
        logger.info(f"{metric}: {score}")


def predict(settings: PredictSettings):
    logger = get_logger()
    path_to_predict_data = settings.predict_data_path
    path_to_model_file = settings.mdl_file
    output_path = settings.output_path
    model = load_model(str(path_to_model_file))
    io_handler = _get_predict_io_handler(
        predict_data_path=path_to_predict_data, output_path=output_path
    )
    predict_data = io_handler.load_predict_data()
    logger.info("Run predict with model:\n%s", model)

    scores = model.predict(predict_data)
    io_handler.store(scores)


def load_model(path_to_model_file: str):
    return load(path_to_model_file)


def _load_train_input(p: Path) -> TrainData:
    if p.is_dir():
        train_input = AnnifHandler(dir=p).load_train_input().train_data
    else:
        train_input = load_tsv_train_input(p)
    return train_input


class PredictTSVIOHandler:
    def __init__(self, predict_data_path: Path, output_path: Optional[Path] = None):
        self._predict_data_path = predict_data_path
        self._output_path = output_path

    def load_predict_data(self):
        return load_predict_input(self._predict_data_path)

    def store(self, data: List[Any]):
        with self._output_path.open("w") as f:
            for d in data:
                f.write(str(d) + "\n")


class PredictAnnifIOHandler:
    def __init__(self, predict_data_path: Path):
        self._annif_handler = AnnifHandler(dir=predict_data_path)
        self._doc_ids = []

    def load_predict_data(self) -> PredictData:
        data = self._annif_handler.load_predict_input()
        self._doc_ids = data.document_ids
        return data.predict_data

    def store(self, data: List[Any]):
        quality_ests = zip(data, self._doc_ids)
        self._annif_handler.store_quality_estimations(quality_ests)


def _get_predict_io_handler(
    predict_data_path: Path, output_path: Optional[Path]
) -> Union[PredictAnnifIOHandler, PredictTSVIOHandler]:
    if predict_data_path.is_dir():
        return PredictAnnifIOHandler(predict_data_path)
    return PredictTSVIOHandler(
        predict_data_path=predict_data_path, output_path=output_path
    )



#  Copyright 2021-2025 ZBW – Leibniz Information Centre for Economics
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from typing import List

import numpy as np
from scipy.sparse import issparse
from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.ensemble import ExtraTreesRegressor
from sklearn.utils.validation import check_is_fitted

from qualle.label_calibration.simple import LabelCalibrator
from qualle.models import Matrix


class MultiCategoryLabelCalibrator(BaseEstimator, RegressorMixin):
    """Label calibrator for multiple distinct categories.
    E.g. to predict the no of labels for different thesauri."""

    def __init__(self, regressor_class=ExtraTreesRegressor, regressor_params=None):
        self.regressor_class = regressor_class
        self.regressor_params = regressor_params or {}

    def fit(self, X: List[str], y: Matrix):
        """
        :param X: list of content, e.g. doc titles
        :param y: matrix where each column corresponds to no of labels for
            respective category
        :return: self
        """
        try:
            no_categories = y.shape[1]
        except IndexError:
            raise ValueError("Number of categories must be greater 0")

        self.calibrators_ = [
            LabelCalibrator(self.regressor_class(**self.regressor_params))
            for _ in range(no_categories)
        ]

        y_is_sparse = issparse(y)
        for i, c in enumerate(self.calibrators_):
            if y_is_sparse:
                col = y.getcol(i).toarray().flatten()
            else:
                col = y[:, i]
            c.fit(X, col)
        return self

    def predict(self, X: List[str]):
        check_is_fitted(self)
        return np.column_stack([c.predict(X) for c in self.calibrators_])



#  Copyright 2021-2025 ZBW  – Leibniz Information Centre for Economics
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import csv
from collections import namedtuple
from pathlib import Path

from qualle.models import TrainData, PredictData

_Data = namedtuple("data", ["docs", "predicted_labels", "scores", "true_labels"])


def load_train_input(p: Path) -> TrainData:
    return _map_to_train_data(_load_input(p))


def load_predict_input(p: Path):
    return _map_to_predict_data(_load_input(p, include_true_labels=False))


def _map_to_train_data(data: _Data) -> TrainData:
    return TrainData(
        predict_data=_map_to_predict_data(data), true_labels=data.true_labels
    )


def _map_to_predict_data(data: _Data) -> PredictData:
    return PredictData(
        docs=data.docs, predicted_labels=data.predicted_labels, scores=data.scores
    )


def _load_input(path_to_tsv: Path, include_true_labels=True) -> _Data:
    docs = []
    pred_labels = []
    true_labels = []
    scores = []

    with open(path_to_tsv, newline="") as csvfile:
        reader = csv.reader(csvfile, delimiter="\t")
        for row in reader:
            docs.append(row[0])
            pred_labels_for_row = []
            scores_for_row = []
            label_score_pairs = row[1].split(",")
            if len(label_score_pairs) > 0 and label_score_pairs[0]:
                for label_score_pair in label_score_pairs:
                    label, score = label_score_pair.split(":")
                    pred_labels_for_row.append(label)
                    scores_for_row.append(float(score))
            pred_labels.append(pred_labels_for_row)
            scores.append(scores_for_row)

            if include_true_labels and len(row) > 2:
                true_labels.append(list(filter(bool, row[2].split(","))))

    return _Data(
        docs=docs, predicted_labels=pred_labels, scores=scores, true_labels=true_labels
    )



#  Copyright 2021-2025 ZBW – Leibniz Information Centre for Economics
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from enum import Enum
from pathlib import Path
from typing import Dict, List, Optional, Union
from typing_extensions import Annotated

from pydantic import (
    model_validator,
    FilePath,
    DirectoryPath,
    TypeAdapter,
    PlainValidator,
    AfterValidator,
)
from pydantic_settings import BaseSettings
from pydantic.networks import AnyUrl
from qualle.features.confidence import ConfidenceFeatures
from qualle.features.text import TextFeatures

# From pydantic v2 onwards, AnyUrl object does not inherit from a string class.
# The following code block performs validation on a pydantic AnyUrl object
# as if it were a string.  Another problem is that a trailing '/' character is also
# appended in pydantic v2 and it is being removed in the code block given below.

AnyUrlAdapter = TypeAdapter(AnyUrl)
HttpUrlStr = Annotated[
    str,
    PlainValidator(lambda x: AnyUrlAdapter.validate_strings(x)),
    AfterValidator(lambda x: str(x).rstrip("/")),
]


FileOrDirPath = Union[FilePath, DirectoryPath]


class RegressorSettings(BaseSettings):
    regressor_class: str
    params: Dict


class SubthesauriLabelCalibrationSettings(BaseSettings):
    thesaurus_file: FilePath
    subthesaurus_type: HttpUrlStr
    concept_type: HttpUrlStr
    concept_type_prefix: HttpUrlStr
    subthesauri: List[HttpUrlStr]
    use_sparse_count_matrix: bool = False


class FeaturesEnum(Enum):
    CONFIDENCE = ConfidenceFeatures
    TEXT = TextFeatures


class TrainSettings(BaseSettings):

    label_calibrator_regressor: RegressorSettings
    quality_estimator_regressor: RegressorSettings

    train_data_path: FileOrDirPath
    output_path: Path

    features: List[FeaturesEnum]

    subthesauri_label_calibration: Optional[SubthesauriLabelCalibrationSettings] = None

    should_debug: bool = False


class EvalSettings(BaseSettings):
    test_data_path: FileOrDirPath
    mdl_file: FilePath


class PredictSettings(BaseSettings):
    predict_data_path: FileOrDirPath
    mdl_file: FilePath
    output_path: Optional[Path] = None

    @model_validator(mode="after")
    def check_output_path_specified_for_input_file(self):
        predict_data_path = self.predict_data_path
        output_path = self.output_path
        if predict_data_path.is_file() and not output_path:
            raise ValueError(
                "output_path has to be specified if predict_data_path "
                "refers to a file"
            )
        return self


class RESTSettings(BaseSettings):
    mdl_file: FilePath
    port: int = 8000
    host: str = "127.0.0.1"



#  Copyright 2021-2025 ZBW – Leibniz Information Centre for Economics
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from dataclasses import dataclass
from typing import List, Union

import numpy as np
from pydantic import model_validator, BaseModel
from scipy.sparse import spmatrix

Labels = List[str]
Documents = List[str]
Scores = List[float]
Matrix = Union[np.ndarray, spmatrix]


class PredictData(BaseModel):
    docs: Documents
    predicted_labels: List[Labels]
    scores: List[Scores]

    @model_validator(mode="after")
    def check_equal_length(self):
        length = None
        for v in self.__dict__.values():
            if length is None:
                length = len(v)
            else:
                if length != len(v):
                    raise ValueError(
                        "docs, predicted_labels and scores "
                        "should have the same length"
                    )
        return self


class TrainData(BaseModel):

    predict_data: PredictData
    true_labels: List[Labels]

    @model_validator(mode="after")
    def check_equal_length(self):
        p_data = self.predict_data
        t_labels = self.true_labels
        if len(p_data.predicted_labels) != len(t_labels):
            raise ValueError("length of true labels and predicted labels do not match")
        return self


@dataclass
class LabelCalibrationData:

    predicted_no_of_labels: np.array
    predicted_labels: List[Labels]



#  Copyright 2021-2025 ZBW – Leibniz Information Centre for Economics
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from enum import Enum
from functools import lru_cache
from typing import List, Optional

import uvicorn
from fastapi import status, FastAPI, Depends, APIRouter
from pydantic.main import BaseModel

from qualle.interface.config import RESTSettings
from qualle.interface.internal import load_model as internal_load_model
from qualle.models import PredictData
from qualle.pipeline import QualityEstimationPipeline

PREDICT_ENDPOINT = "/predict"


class Document(BaseModel):
    content: str
    predicted_labels: List[str]
    scores: List[float]


class Documents(BaseModel):
    documents: List[Document]


class Metric(Enum):
    RECALL = "recall"


class QualityScores(BaseModel):
    name: Metric
    scores: List[float]


class QualityEstimation(BaseModel):

    scores: List[QualityScores]


router = APIRouter()


@lru_cache
def load_model() -> QualityEstimationPipeline:
    settings = RESTSettings()
    return internal_load_model(str(settings.mdl_file))


@router.post(
    PREDICT_ENDPOINT, status_code=status.HTTP_200_OK, response_model=QualityEstimation
)
def predict(
    documents: Documents, qe_pipeline: QualityEstimationPipeline = Depends(load_model)
) -> QualityEstimation:
    predict_data = _map_documents_to_predict_data(documents)
    scores = qe_pipeline.predict(predict_data)
    return QualityEstimation(
        scores=[QualityScores(name=Metric.RECALL, scores=list(scores))]
    )


@router.get("/_up")
def up():
    return True


def _map_documents_to_predict_data(documents: Documents) -> PredictData:
    docs = []
    p_labels = []
    scores = []

    for idx, doc in enumerate(documents.documents):
        docs.append(doc.content)
        p_labels.append(doc.predicted_labels)
        scores.append(doc.scores)
    return PredictData(docs=docs, predicted_labels=p_labels, scores=scores)


def create_app(settings: Optional[RESTSettings] = None):
    settings = settings or RESTSettings()
    app = FastAPI()
    app.include_router(router)
    m = internal_load_model(str(settings.mdl_file))
    app.dependency_overrides[load_model] = lambda: m

    return app


def run(settings: Optional[RESTSettings] = None):
    app = create_app(settings)
    uvicorn.run(app, host=settings.host, port=settings.port)
